@@ -1,30 +1,30 @@
 """This module implements the Scraper component which parses responses and
 extracts information from them"""
 
+from collections import deque
+
 from twisted.python.failure import Failure
 from twisted.internet import defer
 
 from scrapy.utils.defer import defer_result, defer_succeed, parallel, iter_errback
 from scrapy.utils.spider import iterate_spider_output
 from scrapy.utils.misc import load_object
-from scrapy.utils.signal import send_catch_log, send_catch_log_deferred
-from scrapy.exceptions import IgnoreRequest, DropItem
+from scrapy.exceptions import CloseSpider, DropItem
 from scrapy import signals
 from scrapy.http import Request, Response
 from scrapy.item import BaseItem
 from scrapy.core.spidermw import SpiderMiddlewareManager
 from scrapy import log
-from scrapy.stats import stats
 
 
-class SpiderInfo(object):
-    """Object for holding data of the responses being scraped"""
+class Slot(object):
+    """Scraper slot (one per running spider)"""
 
     MIN_RESPONSE_SIZE = 1024
 
     def __init__(self, max_active_size=5000000):
         self.max_active_size = max_active_size
-        self.queue = []
+        self.queue = deque()
         self.active = set()
         self.active_size = 0
         self.itemproc_size = 0
@@ -40,12 +40,12 @@ class SpiderInfo(object):
         return deferred
 
     def next_response_request_deferred(self):
-        response, request, deferred = self.queue.pop(0)
-        self.active.add(response)
+        response, request, deferred = self.queue.popleft()
+        self.active.add(request)
         return response, request, deferred
 
-    def finish_response(self, response):
-        self.active.remove(response)
+    def finish_response(self, response, request):
+        self.active.remove(request)
         if isinstance(response, Response):
             self.active_size -= max(len(response.body), self.MIN_RESPONSE_SIZE)
         else:
@@ -59,58 +59,58 @@ class SpiderInfo(object):
 
 class Scraper(object):
 
-    def __init__(self, engine, settings):
-        self.sites = {}
-        self.spidermw = SpiderMiddlewareManager.from_settings(settings)
-        itemproc_cls = load_object(settings['ITEM_PROCESSOR'])
-        self.itemproc = itemproc_cls.from_settings(settings)
-        self.concurrent_items = settings.getint('CONCURRENT_ITEMS')
-        self.engine = engine
+    def __init__(self, crawler):
+        self.slots = {}
+        self.spidermw = SpiderMiddlewareManager.from_crawler(crawler)
+        itemproc_cls = load_object(crawler.settings['ITEM_PROCESSOR'])
+        self.itemproc = itemproc_cls.from_crawler(crawler)
+        self.concurrent_items = crawler.settings.getint('CONCURRENT_ITEMS')
+        self.crawler = crawler
+        self.signals = crawler.signals
+        self.logformatter = crawler.logformatter
 
     @defer.inlineCallbacks
     def open_spider(self, spider):
         """Open the given spider for scraping and allocate resources for it"""
-        assert spider not in self.sites, "Spider already opened: %s" % spider
-        self.sites[spider] = SpiderInfo()
+        assert spider not in self.slots, "Spider already opened: %s" % spider
+        self.slots[spider] = Slot()
         yield self.itemproc.open_spider(spider)
 
     def close_spider(self, spider):
         """Close a spider being scraped and release its resources"""
-        assert spider in self.sites, "Spider not opened: %s" % spider
-        site = self.sites[spider]
-        site.closing = defer.Deferred()
-        site.closing.addCallback(self.itemproc.close_spider)
-        self._check_if_closing(spider, site)
-        return site.closing
+        assert spider in self.slots, "Spider not opened: %s" % spider
+        slot = self.slots[spider]
+        slot.closing = defer.Deferred()
+        slot.closing.addCallback(self.itemproc.close_spider)
+        self._check_if_closing(spider, slot)
+        return slot.closing
 
     def is_idle(self):
         """Return True if there isn't any more spiders to process"""
-        return not self.sites
+        return not self.slots
 
-    def _check_if_closing(self, spider, site):
-        if site.closing and site.is_idle():
-            del self.sites[spider]
-            site.closing.callback(spider)
+    def _check_if_closing(self, spider, slot):
+        if slot.closing and slot.is_idle():
+            del self.slots[spider]
+            slot.closing.callback(spider)
 
     def enqueue_scrape(self, response, request, spider):
-        site = self.sites.get(spider, None)
-        if site is None:
-            return
-        dfd = site.add_response_request(response, request)
+        slot = self.slots[spider]
+        dfd = slot.add_response_request(response, request)
         def finish_scraping(_):
-            site.finish_response(response)
-            self._check_if_closing(spider, site)
-            self._scrape_next(spider, site)
+            slot.finish_response(response, request)
+            self._check_if_closing(spider, slot)
+            self._scrape_next(spider, slot)
             return _
         dfd.addBoth(finish_scraping)
         dfd.addErrback(log.err, 'Scraper bug processing %s' % request, \
             spider=spider)
-        self._scrape_next(spider, site)
+        self._scrape_next(spider, slot)
         return dfd
 
-    def _scrape_next(self, spider, site):
-        while site.queue:
-            response, request, deferred = site.next_response_request_deferred()
+    def _scrape_next(self, spider, slot):
+        while slot.queue:
+            response, request, deferred = slot.next_response_request_deferred()
             self._scrape(response, request, spider).chainDeferred(deferred)
 
     def _scrape(self, response, request, spider):
@@ -119,7 +119,7 @@ class Scraper(object):
         assert isinstance(response, (Response, Failure))
 
         dfd = self._scrape2(response, request, spider) # returns spiders processed output
-        dfd.addErrback(self.handle_spider_error, request, spider)
+        dfd.addErrback(self.handle_spider_error, request, response, spider)
         dfd.addCallback(self.handle_spider_output, request, response, spider)
         return dfd
 
@@ -132,26 +132,30 @@ class Scraper(object):
         else:
             # FIXME: don't ignore errors in spider middleware
             dfd = self.call_spider(request_result, request, spider)
-            return dfd.addErrback(self._check_propagated_failure, \
+            return dfd.addErrback(self._log_download_errors, \
                 request_result, request, spider)
 
     def call_spider(self, result, request, spider):
+        result.request = request
         dfd = defer_result(result)
         dfd.addCallbacks(request.callback or spider.parse, request.errback)
         return dfd.addCallback(iterate_spider_output)
 
-    def handle_spider_error(self, _failure, request, spider, propagated_failure=None):
-        referer = request.headers.get('Referer', None)
-        msg = "Spider error processing <%s> (referer: <%s>)" % \
-            (request.url, referer)
-        log.err(_failure, msg, spider=spider)
-        stats.inc_value("spider_exceptions/%s" % _failure.value.__class__.__name__, \
+    def handle_spider_error(self, _failure, request, response, spider):
+        exc = _failure.value
+        if isinstance(exc, CloseSpider):
+            self.crawler.engine.close_spider(spider, exc.reason or 'cancelled')
+            return
+        log.err(_failure, "Spider error processing %s" % request, spider=spider)
+        self.signals.send_catch_log(signal=signals.spider_error, failure=_failure, response=response, \
+            spider=spider)
+        self.crawler.stats.inc_value("spider_exceptions/%s" % _failure.value.__class__.__name__, \
             spider=spider)
 
     def handle_spider_output(self, result, request, response, spider):
         if not result:
             return defer_succeed(None)
-        it = iter_errback(result, self.handle_spider_error, request, spider)
+        it = iter_errback(result, self.handle_spider_error, request, response, spider)
         dfd = parallel(it, self.concurrent_items,
             self._process_spidermw_output, request, response, spider)
         return dfd
@@ -161,54 +165,50 @@ class Scraper(object):
         from the given spider
         """
         if isinstance(output, Request):
-            send_catch_log(signal=signals.request_received, request=output, \
+            self.signals.send_catch_log(signal=signals.request_received, request=output, \
                 spider=spider)
-            self.engine.crawl(request=output, spider=spider)
+            self.crawler.engine.crawl(request=output, spider=spider)
         elif isinstance(output, BaseItem):
-            log.msg(log.formatter.scraped(output, request, response, spider), \
-                level=log.DEBUG, spider=spider)
-            self.sites[spider].itemproc_size += 1
-            dfd = send_catch_log_deferred(signal=signals.item_scraped, \
-                item=output, spider=spider, response=response)
-            dfd.addBoth(lambda _: self.itemproc.process_item(output, spider))
-            dfd.addBoth(self._itemproc_finished, output, spider)
+            self.slots[spider].itemproc_size += 1
+            dfd = self.itemproc.process_item(output, spider)
+            dfd.addBoth(self._itemproc_finished, output, response, spider)
             return dfd
         elif output is None:
             pass
         else:
-            log.msg("Spider must return Request, BaseItem or None, got %r in %s" % \
-                (type(output).__name__, request), log.ERROR, spider=spider)
+            typename = type(output).__name__
+            log.msg(format='Spider must return Request, BaseItem or None, '
+                           'got %(typename)r in %(request)s',
+                    level=log.ERROR, spider=spider, request=request, typename=typename)
 
-    def _check_propagated_failure(self, spider_failure, propagated_failure, request, spider):
-        """Log and silence the bugs raised outside of spiders, but still allow
-        spiders to be notified about general failures while downloading spider
-        generated requests
+    def _log_download_errors(self, spider_failure, download_failure, request, spider):
+        """Log and silence errors that come from the engine (typically download
+        errors that got propagated thru here)
         """
-        # ignored requests are commonly propagated exceptions safes to be silenced
-        if isinstance(spider_failure.value, IgnoreRequest):
+        if spider_failure is download_failure:
+            errmsg = spider_failure.getErrorMessage()
+            if errmsg:
+                log.msg(format='Error downloading %(request)s: %(errmsg)s',
+                        level=log.ERROR, spider=spider, request=request, errmsg=errmsg)
             return
-        elif spider_failure is propagated_failure:
-            log.err(spider_failure, 'Unhandled error propagated to spider', \
-                spider=spider)
-            return # stop propagating this error
-        else:
-            return spider_failure # exceptions raised in the spider code
+        return spider_failure
 
-    def _itemproc_finished(self, output, item, spider):
+    def _itemproc_finished(self, output, item, response, spider):
         """ItemProcessor finished for the given ``item`` and returned ``output``
         """
-        self.sites[spider].itemproc_size -= 1
+        self.slots[spider].itemproc_size -= 1
         if isinstance(output, Failure):
             ex = output.value
             if isinstance(ex, DropItem):
-                log.msg(log.formatter.dropped(item, ex, spider), \
-                    level=log.WARNING, spider=spider)
-                return send_catch_log_deferred(signal=signals.item_dropped, \
+                logkws = self.logformatter.dropped(item, ex, response, spider)
+                log.msg(level=log.WARNING, spider=spider, **logkws)
+                return self.signals.send_catch_log_deferred(signal=signals.item_dropped, \
                     item=item, spider=spider, exception=output.value)
             else:
-                log.err(output, 'Error processing %s' % item, spider=spider)
+                log.err(output, 'Error processing %(item)s', item=item, spider=spider)
         else:
-            log.msg(log.formatter.passed(output, spider), log.INFO, spider=spider)
-            return send_catch_log_deferred(signal=signals.item_passed, \
-                item=item, spider=spider, output=output)
+            logkws = self.logformatter.scraped(output, response, spider)
+            log.msg(level=log.DEBUG, spider=spider, **logkws)
+            return self.signals.send_catch_log_deferred(signal=signals.item_scraped, \
+                item=output, response=response, spider=spider)
 
