@@ -5,8 +5,6 @@ These exceptions are documented in docs/topics/exceptions.rst. Please don't add
 new exceptions here without documenting them there.
 """
 
-from scrapy import log
-
 # Internal
 
 class NotConfigured(Exception):
@@ -18,17 +16,16 @@ class NotConfigured(Exception):
 class IgnoreRequest(Exception):
     """Indicates a decision was made not to process a request"""
 
-    def __init__(self, msg='', level=log.ERROR):
-        self.msg = msg
-        self.level = level
-
-    def __str__(self):
-        return self.msg
-
 class DontCloseSpider(Exception):
     """Request the spider not to be closed yet"""
     pass
 
+class CloseSpider(Exception):
+    """Raise this from callbacks to request the spider to be closed"""
+
+    def __init__(self, reason='cancelled'):
+        self.reason = reason
+
 # Items
 
 class DropItem(Exception):
@@ -46,3 +43,16 @@ class UsageError(Exception):
     def __init__(self, *a, **kw):
         self.print_help = kw.pop('print_help', True)
         super(UsageError, self).__init__(*a, **kw)
+
+class ScrapyDeprecationWarning(Warning):
+    """Warning category for deprecated features, since the default
+    DeprecationWarning is silenced on Python 2.7+
+    """
+    pass
+
+class ContractFail(AssertionError):
+    """Error raised in case of a failing contract"""
+    
+    def format(self, method):
+        return '[FAILED] %s:%s\n>>> %s\n' % \
+            (method.im_class.name, method.__name__, self)
